/****************************************************************************
**  CUBE        http://www.scalasca.org/                                   **
*****************************************************************************
**  Copyright (c) 2015-2023                                                **
**  Forschungszentrum Juelich GmbH, Juelich Supercomputing Centre          **
**                                                                         **
**  This software may be modified and distributed under the terms of       **
**  a BSD-style license.  See the COPYING file in the package base         **
**  directory for details.                                                 **
****************************************************************************/


#ifndef CUBELIB_POP_HYBRID_COMMUNICATION_EFFICENCY_TEST_ADD_H
#define CUBELIB_POP_HYBRID_COMMUNICATION_EFFICENCY_TEST_ADD_H

#include <string>
#include "PerformanceTest.h"
#include "POPHybridSerialisationTestAdd.h"
#include "POPHybridTransferTestAdd.h"

namespace hybaddanalysis
{
class POPHybridProcessEfficiencyTestAdd;
class POPHybridCommunicationEfficiencyTestAdd : public popcalculation::PerformanceTest
{
    friend class POPHybridProcessEfficiencyTestAdd;
private:
    cube::Metric*                  pop_commeff;
    cube::Metric*                  max_omp_serial_comp_time;
    cube::Metric*                  max_runtime;
    POPHybridSerialisationTestAdd* pop_ser;
    POPHybridTransferTestAdd*      pop_transeff;
    bool                           scout_cubex;
    bool                           original_scout_cubex;


    cube::list_of_metrics lmax_omp_ser_comp_time;

    double
    calculateForScout( const cube::list_of_cnodes& cnodes  ) const;

    void
    add_comm_eff( cube::CubeProxy* ) const;



protected:

    const std::string&
    getCommentText() const;

    inline
    virtual
    std::string
    getHelpUrl() const
    {
        std::string active_text =
            "MPI Communication Efficiency (CommE) can be evaluated directly by following formula:\n"
            "\tComm E= (max(time in OpenMP + serial computation time)) / max(runtime)\n"
            "CommE identifies when code is inefficient because it spends a large amount \n"
            "of time communicating rather than performing useful computations. \n"
            "CommE is composed of two additional metrics that reflect two causes of excessive time within communication:\n"
            "\tProcesses waiting at communication points for other processes to arrive (i.e. serialisation) \n"
            "\tProcesses transferring large amounts of data relative to the network capacity \n"
            "These are measured using Serialisation Efficiency and Transfer Efficiency. \n"
            "Combination of these two sub-metrics gives us Communication Efficiency: \n"
            "\tCommE = Serialisation Efficiency x Transfer Efficiency \n"
            "To obtain these two sub-metrics we need to perform Scalasca trace analysis \n"
            "which identifies serialisations and inefficient communication patterns. ";

        std::string not_active_text =
            "MPI Communication Efficiency (CommE) metric is a basic POP metric and is available for every Score-P/Scalasca measurement. \n"
            "If Cube Report was produced by another tool than Score-P/Scalasca, it might have missing metric Time.\n"
            "In this case POP analysis is not possible.";

        return active_text;
    }

    virtual
    void
    adjustForTest( cube::CubeProxy* cube ) const;


public:
    POPHybridCommunicationEfficiencyTestAdd(
        cube::CubeProxy*,
        POPHybridSerialisationTestAdd* pop_ser,
        POPHybridTransferTestAdd*      pop_transef );

    virtual
    ~POPHybridCommunicationEfficiencyTestAdd()
    {
    };

    virtual
    double
    analyze( const cube::list_of_cnodes& cnodes,
             cube::LocationGroup*        _lg = nullptr ) const;

// ------ overview tests ---------

    void
    applyCnode( const cube::list_of_cnodes& cnodes,
                const bool                  direct_calculation = false  );

    bool
    isActive() const;

    bool
    isIssue() const;

    virtual
    std::list<PerformanceTest*>
    getPrereqs()
    {
        std::list<PerformanceTest*> prereqs;
        if ( pop_ser != nullptr && pop_transeff != nullptr )
        {
            prereqs.push_back( pop_transeff );
            prereqs.push_back( pop_ser );
        }
        return prereqs;
    }
};
};
#endif // ADVISER_RATING_WIDGET_H
