/****************************************************************************
**  CUBE        http://www.scalasca.org/                                   **
*****************************************************************************
**  Copyright (c) 2015-2023                                                **
**  Forschungszentrum Juelich GmbH, Juelich Supercomputing Centre          **
**                                                                         **
**  This software may be modified and distributed under the terms of       **
**  a BSD-style license.  See the COPYING file in the package base         **
**  directory for details.                                                 **
****************************************************************************/


#ifndef CUBELIB_POP_HYBRID_PROCESS_EFFICIENCY_TEST_ADD_H
#define CUBELIB_POP_HYBRID_PROCESS_EFFICIENCY_TEST_ADD_H

#include <string>
#include "PerformanceTest.h"
#include "POPHybridImbalanceTestAdd.h"
#include "POPHybridCommunicationEfficiencyTestAdd.h"

namespace hybaddanalysis
{
class POPHybridParallelEfficiencyTestAdd;
class POPHybridProcessEfficiencyTestAdd : public popcalculation::PerformanceTest
{
    friend class POPHybridParallelEfficiencyTestAdd;
private:


    POPHybridImbalanceTestAdd*               pop_lb;
    POPHybridCommunicationEfficiencyTestAdd* pop_commeff;


    cube::Metric* max_runtime;
    cube::Metric* pop_avg_omp;
    cube::Metric* pop_avg_ser_comp;

    cube::list_of_metrics max_runtime_metrics;
    cube::list_of_metrics lavg_omp_metrics;
    cube::list_of_metrics lavg_ser_metrics;

    void
    calculate();

    void
    calculate( const cube::list_of_cnodes& );


protected:

    virtual
    const std::string&
    getCommentText() const;

    inline
    virtual
    std::string
    getHelpUrl() const
    {
        std::string active_text =
            "Process Efficiency completely ignores thread behavior, and evaluates process utilization via two components:\n"
            "\tWorkload across processes\n"
            "\tCommunication across processes\n"
            "These two can be measured with Computation Load Balance and Communication Efficiency respectively. \n"
            "Process Efficiency can be computed directly or as a sum of these two sub-metrics minus one:\n"
            "\tProcess Efficiency=(avg(time in OpenMP) + avg(serial computation))/ max(runtime) \n"
            "\t= Computation Load Balance + MPI Communication Efficiency - 1 \n"
            "Where average time in OpenMP and average serial computation are computed as weighted arithmetic mean. \n"
            "If number of threads is equal across processes average time in OpenMP and average serial computation \n"
            "can be computed as ordinary arithmetic mean. ";
        std::string not_active_text =
            "Process Efficiency metric is a basic POP metric and is available for every Score-P/Scalasca measurement. \n"
            "If Cube Report was produced by another tool than Score-P/Scalasca, it might have missing metric Time.\n"
            "In this case POP analysis is not possible.";

        return active_text;
    }

    virtual
    void
    adjustForTest( cube::CubeProxy* cube ) const;


public:
    POPHybridProcessEfficiencyTestAdd(      cube::CubeProxy*,
                                            POPHybridImbalanceTestAdd*               pop_lb,
                                            POPHybridCommunicationEfficiencyTestAdd* pop_commeff );

    virtual
    ~POPHybridProcessEfficiencyTestAdd()
    {
    };
// ------ overview tests ---------

    void
    applyCnode( const cube::list_of_cnodes& cnodes,
                const bool                  direct_calculation = false   );



    bool
    isActive() const;

    bool
    isIssue() const;

    virtual
    std::list<PerformanceTest*>
    getPrereqs()
    {
        std::list<PerformanceTest*> prereqs;
        if ( pop_lb != nullptr && pop_commeff != nullptr )
        {
            prereqs.push_back( pop_lb );
            prereqs.push_back( pop_commeff );
        }
        return prereqs;
    }
};
};
#endif // ADVISER_RATING_WIDGET_H
